#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
CoAP resources to deal with Property interactions.
"""

import json
import logging

import aiocoap
import aiocoap.error
import aiocoap.resource
import tornado.gen

from wotpy.protocols.coap.resources.utils import parse_request_opt_query

JSON_CONTENT_FORMAT = 50


@tornado.gen.coroutine
def _build_property_value_response(thing_property):
    """Reads the current property value and builds
    the CoAP response containing said value."""

    value = yield thing_property.read()
    payload = json.dumps({"value": value}).encode("utf-8")
    response = aiocoap.Message(code=aiocoap.Code.CONTENT, payload=payload)
    response.opt.content_format = JSON_CONTENT_FORMAT
    raise tornado.gen.Return(response)


def get_thing_property(server, request):
    """Takes a CoAP request and returns the Thing Property
    identified by the request arguments."""

    query = parse_request_opt_query(request)
    url_name_thing = query.get("thing")
    url_name_prop = query.get("name")

    if not url_name_thing or not url_name_prop:
        raise aiocoap.error.BadRequest("Missing query arguments")

    exposed_thing = server.exposed_thing_set.find_by_thing_id(url_name_thing)

    if not exposed_thing:
        raise aiocoap.error.NotFound("Thing not found")

    try:
        return next(
            exposed_thing.properties[key] for key in exposed_thing.properties
            if exposed_thing.properties[key].url_name == url_name_prop)
    except StopIteration:
        raise aiocoap.error.NotFound("Property not found")


class PropertyReadWriteResource(aiocoap.resource.Resource):
    """CoAP resource to handle Property reads and writes."""

    def __init__(self, server):
        super(PropertyReadWriteResource, self).__init__()
        self._server = server

    @tornado.gen.coroutine
    def render_get(self, request):
        """Returns a CoAP response with the current property value."""

        thing_property = get_thing_property(self._server, request)
        response = yield _build_property_value_response(thing_property)
        raise tornado.gen.Return(response)

    @tornado.gen.coroutine
    def render_post(self, request):
        """Updates the property with the value retrieved from the CoAP request payload."""

        thing_property = get_thing_property(self._server, request)
        request_payload = json.loads(request.payload)

        if "value" not in request_payload:
            raise aiocoap.error.BadRequest()

        yield thing_property.write(request_payload.get("value"))
        response = aiocoap.Message(code=aiocoap.Code.CHANGED)

        raise tornado.gen.Return(response)


class PropertyObservableResource(aiocoap.resource.ObservableResource):
    """CoAP resource to observe property updates."""

    def __init__(self, server):
        super(PropertyObservableResource, self).__init__()
        self._server = server
        self._subscription = None
        self._logr = logging.getLogger(__name__)

    @tornado.gen.coroutine
    def add_observation(self, request, server_observation):
        """Method that decides whether to add a new observer.
        A new observer is added for each GET request."""

        if request.code.name != aiocoap.Code.GET.name:
            return

        try:
            thing_property = get_thing_property(self._server, request)
        except aiocoap.error.Error:
            return

        # noinspection PyUnusedLocal
        def on_next(item):
            server_observation.trigger()

        def on_error(err):
            self._logr.warning("Error on subscription to {}: {}".format(thing_property, err))

        subscription = thing_property.subscribe(on_next=on_next, on_error=on_error)

        def cancellation_cb():
            self._logr.debug("Disposing of subscription to: {}".format(thing_property))
            subscription.dispose()

        server_observation.accept(cancellation_cb)

    # noinspection PyUnusedLocal
    @tornado.gen.coroutine
    def render_get(self, request):
        """Returns a CoAP response with the current property value."""

        thing_property = get_thing_property(self._server, request)
        response = yield _build_property_value_response(thing_property)
        raise tornado.gen.Return(response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

TD_EXAMPLE = {
    "id": "urn:dev:wot:com:example:servient:lamp",
    "name": "MyLampThing",
    "description": "MyLampThing uses JSON-LD 1.1 serialization",
    "security": [{"scheme": "nosec"}],
    "properties": {
        "status": {
            "description": "Shows the current status of the lamp",
            "type": "string",
            "forms": [{
                "href": "coaps://mylamp.example.com/status"
            }]
        }
    },
    "actions": {
        "toggle": {
            "description": "Turn on or off the lamp",
            "forms": [{
                "href": "coaps://mylamp.example.com/toggle"
            }]
        }
    },
    "events": {
        "overheating": {
            "description": "Lamp reaches a critical temperature (overheating)",
            "data": {"type": "string"},
            "forms": [{
                "href": "coaps://mylamp.example.com/oh"
            }]
        }
    }
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Wrapper class for dictionaries to represent Thing filters.
"""

from wotpy.wot.dictionaries.base import WotBaseDict
from wotpy.wot.enums import DiscoveryMethod


class ThingFilterDict(WotBaseDict):
    """The ThingFilter dictionary that represents the
    constraints for discovering Things as key-value pairs."""

    class Meta:
        fields = {
            "method",
            "url",
            "query",
            "fragment"
        }

        defaults = {
            "method": DiscoveryMethod.ANY
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import uuid

import six
import tornado.gen
import tornado.ioloop
from faker import Faker
from rx.concurrency import IOLoopScheduler
from tornado.concurrent import Future

from tests.utils import find_free_port, run_test_coroutine
from wotpy.protocols.http.client import HTTPClient
from wotpy.protocols.http.server import HTTPServer
from wotpy.protocols.ws.client import WebsocketClient
from wotpy.protocols.ws.server import WebsocketServer
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription


def _test_property_change_events(exposed_thing, subscribe_func):
    """Helper function to test client subscriptions to property change events."""

    @tornado.gen.coroutine
    def test_coroutine():
        td = ThingDescription.from_thing(exposed_thing.thing)
        prop_name = next(six.iterkeys(td.properties))

        future_conn = Future()
        future_change = Future()

        prop_value = Faker().sentence()

        def on_next(ev):
            if not future_conn.done():
                future_conn.set_result(True)
                return

            if ev.data.value == prop_value:
                future_change.set_result(True)

        subscription = subscribe_func(prop_name, on_next)

        while not future_conn.done():
            yield tornado.gen.sleep(0)
            yield exposed_thing.write_property(prop_name, Faker().sentence())

        yield exposed_thing.write_property(prop_name, prop_value)

        yield future_change

        assert future_change.result()

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def _test_event_emission_events(exposed_thing, subscribe_func):
    """Helper function to test client subscription to event emissions."""

    @tornado.gen.coroutine
    def test_coroutine():
        td = ThingDescription.from_thing(exposed_thing.thing)
        event_name = next(six.iterkeys(td.events))

        future_conn = Future()
        future_event = Future()

        payload = Faker().sentence()

        def on_next(ev):
            if not future_conn.done():
                future_conn.set_result(True)
                return

            if ev.data == payload:
                future_event.set_result(True)

        subscription = subscribe_func(event_name, on_next)

        while not future_conn.done():
            yield tornado.gen.sleep(0)
            exposed_thing.emit_event(event_name, Faker().sentence())

        exposed_thing.emit_event(event_name, payload)

        yield future_event

        assert future_event.result()

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def test_thing_template_getters(consumed_exposed_pair):
    """ThingTemplate properties can be accessed from the ConsumedThing."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    thing_template = consumed_thing.td.to_thing_fragment()

    assert consumed_thing.id == thing_template.id
    assert consumed_thing.name == thing_template.name
    assert consumed_thing.description == thing_template.description


def test_read_property(consumed_exposed_pair):
    """A ConsumedThing is able to read properties."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = next(six.iterkeys(consumed_thing.td.properties))

        result_exposed = yield exposed_thing.read_property(prop_name)
        result_consumed = yield consumed_thing.read_property(prop_name)

        assert result_consumed == result_exposed

    run_test_coroutine(test_coroutine)


def test_write_property(consumed_exposed_pair):
    """A ConsumedThing is able to write properties."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = next(six.iterkeys(consumed_thing.td.properties))

        val_01 = Faker().sentence()
        val_02 = Faker().sentence()

        yield exposed_thing.write_property(prop_name, val_01)
        value = yield exposed_thing.read_property(prop_name)

        assert value == val_01

        yield consumed_thing.write_property(prop_name, val_02)
        value = yield exposed_thing.read_property(prop_name)

        assert value == val_02

    run_test_coroutine(test_coroutine)


def test_invoke_action(consumed_exposed_pair):
    """A ConsumedThing is able to invoke actions."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = next(six.iterkeys(consumed_thing.td.actions))

        input_value = Faker().pystr()
        result = yield consumed_thing.invoke_action(action_name, input_value)
        result_expected = yield exposed_thing.invoke_action(action_name, input_value)

        assert result == result_expected

    run_test_coroutine(test_coroutine)


def test_on_event(consumed_exposed_pair):
    """A ConsumedThing is able to observe events."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    def subscribe_func(event_name, on_next):
        observable = consumed_thing.on_event(event_name)
        return observable.subscribe_on(IOLoopScheduler()).subscribe(on_next)

    _test_event_emission_events(exposed_thing, subscribe_func)


def test_on_property_change(consumed_exposed_pair):
    """A ConsumedThing is able to observe property updates."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    def subscribe_func(prop_name, on_next):
        observable = consumed_thing.on_property_change(prop_name)
        return observable.subscribe_on(IOLoopScheduler()).subscribe(on_next)

    _test_property_change_events(exposed_thing, subscribe_func)


def test_thing_property_get(consumed_exposed_pair):
    """Property values can be retrieved on ConsumedThings using the map-like interface."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = next(six.iterkeys(consumed_thing.td.properties))

        result_exposed = yield exposed_thing.read_property(prop_name)
        result_consumed = yield consumed_thing.properties[prop_name].read()

        assert result_consumed == result_exposed

    run_test_coroutine(test_coroutine)


def test_thing_property_set(consumed_exposed_pair):
    """Property values can be updated on ConsumedThings using the map-like interface."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = next(six.iterkeys(consumed_thing.td.properties))
        updated_value = Faker().sentence()
        curr_value = yield exposed_thing.read_property(prop_name)

        assert consumed_thing.td.properties[prop_name].writable
        assert curr_value != updated_value

        yield consumed_thing.properties[prop_name].write(updated_value)
        result_exposed = yield exposed_thing.read_property(prop_name)

        assert result_exposed == updated_value

    run_test_coroutine(test_coroutine)


def test_thing_property_subscribe(consumed_exposed_pair):
    """Property updates can be observed on ConsumedThings using the map-like interface."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    def subscribe_func(prop_name, on_next):
        return consumed_thing.properties[prop_name].subscribe(on_next)

    _test_property_change_events(exposed_thing, subscribe_func)


def test_thing_property_getters(consumed_exposed_pair):
    """ThingProperty retrieved from ConsumedThing expose the attributes
    from the Interaction, InteractionFragment and PropertyFragment interfaces."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = next(six.iterkeys(consumed_thing.td.properties))
        thing_prop_con = consumed_thing.properties[prop_name]
        thing_prop_exp = exposed_thing.properties[prop_name]

        assert len(thing_prop_con.forms) == 0
        assert thing_prop_con.title == thing_prop_exp.title
        assert thing_prop_con.description == thing_prop_exp.description
        assert thing_prop_con.observable == thing_prop_exp.observable
        assert thing_prop_con.type == thing_prop_exp.type

    run_test_coroutine(test_coroutine)


def test_thing_action_run(consumed_exposed_pair):
    """Actions can be invoked on ConsumedThings using the map-like interface."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = next(six.iterkeys(consumed_thing.td.actions))

        input_value = Faker().pystr()
        result = yield consumed_thing.actions[action_name].invoke(input_value)
        result_expected = yield exposed_thing.invoke_action(action_name, input_value)

        assert result == result_expected

    run_test_coroutine(test_coroutine)


def test_thing_action_getters(consumed_exposed_pair):
    """ThingAction retrieved from ConsumedThing expose the attributes
    from the Interaction, InteractionFragment and ActionFragment interfaces."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = next(six.iterkeys(consumed_thing.td.actions))
        thing_action_con = consumed_thing.actions[action_name]
        thing_action_exp = exposed_thing.actions[action_name]

        assert len(thing_action_con.forms) == 0
        assert thing_action_con.title == thing_action_exp.title
        assert thing_action_con.description == thing_action_exp.description
        assert thing_action_con.input.type == thing_action_exp.input.type
        assert thing_action_con.output.type == thing_action_exp.output.type

    run_test_coroutine(test_coroutine)


def test_thing_event_subscribe(consumed_exposed_pair):
    """Property updates can be observed on ConsumedThings using the map-like interface."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    def subscribe_func(event_name, on_next):
        return consumed_thing.events[event_name].subscribe(on_next)

    _test_event_emission_events(exposed_thing, subscribe_func)


def test_thing_event_getters(consumed_exposed_pair):
    """ThingEvent retrieved from ConsumedThing expose the attributes
    from the Interaction, InteractionFragment and EventFragment interfaces."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")
    exposed_thing = consumed_exposed_pair.pop("exposed_thing")

    @tornado.gen.coroutine
    def test_coroutine():
        event_name = next(six.iterkeys(consumed_thing.td.events))
        thing_event_con = consumed_thing.events[event_name]
        thing_event_exp = exposed_thing.events[event_name]

        assert len(thing_event_con.forms) == 0
        assert thing_event_con.title == thing_event_exp.title
        assert thing_event_con.description == thing_event_exp.description
        assert thing_event_con.data.type == thing_event_exp.data.type

    run_test_coroutine(test_coroutine)


def test_thing_interaction_dict_behaviour(consumed_exposed_pair):
    """The Interactions dict-like interface of a ConsumedThing behaves like a dict."""

    consumed_thing = consumed_exposed_pair.pop("consumed_thing")

    prop_name = next((key for key in consumed_thing.properties), None)

    assert prop_name
    assert len(consumed_thing.properties) > 0
    assert prop_name in consumed_thing.properties


def test_consumed_client_protocols_preference():
    """The Servient selects different protocol clients to consume Things
    depending on the protocol choices displayed on the Thing Description."""

    servient = Servient(catalogue_port=None)

    @tornado.gen.coroutine
    def servient_start():
        raise tornado.gen.Return((yield servient.start()))

    @tornado.gen.coroutine
    def servient_shutdown():
        yield servient.shutdown()

    http_port = find_free_port()
    http_server = HTTPServer(port=http_port)

    servient.add_server(http_server)

    ws_port = find_free_port()
    ws_server = WebsocketServer(port=ws_port)

    servient.add_server(ws_server)

    client_server_map = {
        HTTPClient: http_server,
        WebsocketClient: ws_server
    }

    wot = tornado.ioloop.IOLoop.current().run_sync(servient_start)

    prop_name = uuid.uuid4().hex

    td_produce = ThingDescription({
        "id": uuid.uuid4().urn,
        "name": uuid.uuid4().hex,
        "properties": {
            prop_name: {
                "observable": True,
                "type": "string"
            }
        }
    })

    exposed_thing = wot.produce(td_produce.to_str())
    exposed_thing.expose()

    td_forms_all = ThingDescription.from_thing(exposed_thing.thing)

    client_01 = servient.select_client(td_forms_all, prop_name)
    client_01_class = client_01.__class__

    assert client_01_class in six.iterkeys(client_server_map)

    tornado.ioloop.IOLoop.current().run_sync(servient_shutdown)
    servient.remove_server(client_server_map[client_01_class].protocol)
    tornado.ioloop.IOLoop.current().run_sync(servient_start)

    td_forms_removed = ThingDescription.from_thing(exposed_thing.thing)

    client_02 = servient.select_client(td_forms_removed, prop_name)
    client_02_class = client_02.__class__

    assert client_02_class != client_01_class
    assert client_02_class in six.iterkeys(client_server_map)

    tornado.ioloop.IOLoop.current().run_sync(servient_shutdown)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Service discovery based on Multicast DNS and DNS-SD (Bonjour, Avahi).
"""

import socket
import threading
import time
from functools import partial
from typing import cast

import six
import tornado.gen
import tornado.ioloop
import tornado.locks
from six.moves import queue
from slugify import slugify
from zeroconf import Zeroconf, ServiceStateChange, ServiceInfo, ServiceBrowser

from wotpy.utils.utils import get_main_ipv4_address

ITER_WAIT = 0.1
TASK_REGISTER = 'register'
TASK_UNREGISTER = 'unregister'


def build_servient_service_info(servient, address=None, instance_name=None):
    """Takes a Servient and optional IP address and builds the
    zeroconf ServiceInfo that describes the WoT Servient service."""

    address = address if address else get_main_ipv4_address()
    servient_urlname = slugify(servient.hostname)
    instance_name = instance_name if instance_name else servient_urlname
    servient_fqdn = "{}.".format(servient.hostname.strip('.'))
    server_default = "{}.local.".format(servient_urlname)
    server = servient_fqdn if servient_fqdn.endswith('.local.') else server_default

    return ServiceInfo(
        DNSSDDiscoveryService.WOT_SERVICE_TYPE,
        "{}.{}".format(instance_name, DNSSDDiscoveryService.WOT_SERVICE_TYPE),
        port=servient.catalogue_port,
        address=socket.inet_aton(address),
        properties={},
        server=server)


def _start_zeroconf(close_event, services, services_lock, register_queue, address):
    """Starts the zeroconf mDNS service.
    Listens to the register tasks queue and starts browsing for WoT Servient services."""

    address = address if address else get_main_ipv4_address()
    zeroconf = Zeroconf()
    registered = []
    registered_lock = threading.Lock()

    def _on_service_change(*args, **kwargs):
        """Callback for each time a WoT Servient service is added or removed from the link."""

        service_type = kwargs.pop('service_type')
        name = kwargs.pop('name')
        state_change = kwargs.pop('state_change')

        with registered_lock:
            is_local = any(item.name == name for item in registered)

        if is_local:
            return

        def _add_result():
            info = zeroconf.get_service_info(service_type, name)
            info_addr = socket.inet_ntoa(cast(bytes, info.address))
            info_port = cast(int, info.port)

            with services_lock:
                services[name] = (info_addr, info_port)

        def _remove_result():
            with services_lock:
                try:
                    services.pop(name)
                except KeyError:
                    pass

        change_handler_map = {
            ServiceStateChange.Added: _add_result,
            ServiceStateChange.Removed: _remove_result
        }

        change_handler_map[state_change]()

    def _register(task):
        """Registers a new WoT Servient service."""

        servient = task['servient']
        done = task['done']
        instance_name = task['instance_name']

        try:
            if not servient.catalogue_port:
                return

            info = build_servient_service_info(
                servient,
                address=address,
                instance_name=instance_name)

            with registered_lock:
                registered.append(info)

            zeroconf.register_service(info)
        finally:
            done.set()

    def _unregister(task):
        """Unregisters a WoT Servient service."""

        servient = task['servient']
        done = task['done']
        instance_name = task['instance_name']

        try:
            info = build_servient_service_info(
                servient,
                address=address,
                instance_name=instance_name)

            with registered_lock:
                is_registered = any(val == info for val in registered)

            if not is_registered:
                return

            zeroconf.unregister_service(info)

            with registered_lock:
                registered.remove(info)
        finally:
            done.set()

    def _main_loop():
        """Main Zeroconf service loop that starts browsing for WoT
        services and processes the register tasks queue."""

        browser = None

        try:
            browser = ServiceBrowser(
                zeroconf,
                DNSSDDiscoveryService.WOT_SERVICE_TYPE,
                handlers=[_on_service_change])

            while not close_event.is_set():
                try:
                    register_task = register_queue.get_nowait()

                    task_handler_map = {
                        TASK_REGISTER: _register,
                        TASK_UNREGISTER: _unregister
                    }

                    task_handler_map[register_task['type']](register_task)
                except queue.Empty:
                    pass

                close_event.wait(ITER_WAIT)
        finally:
            if browser is not None:
                browser.cancel()

            with registered_lock:
                for serv_info in registered:
                    zeroconf.unregister_service(serv_info)

            zeroconf.close()

    _main_loop()


class DNSSDDiscoveryService(object):
    """Manages a DNS Service Discovery service (based on Multicast DNS)
    that is run on a separate thread (on a loop executor) to discover
    link-local WoT Servients and expose its own."""

    WOT_SERVICE_TYPE = "_wot-servient._tcp.local."

    def __init__(self, address=None):
        self._address = address
        self._zeroconf_loop_thread = None
        self._close_event = threading.Event()
        self._register_queue = None
        self._lock = tornado.locks.Lock()
        self._loop = tornado.ioloop.IOLoop.current()
        self._services = None
        self._services_lock = threading.Lock()

    @property
    def is_running(self):
        """Returns True if the mDNS service is currently running."""

        return self._zeroconf_loop_thread is not None

    @tornado.gen.coroutine
    def start(self):
        """Starts the DNS-SD thread on a loop executor."""

        with (yield self._lock.acquire()):
            if self._zeroconf_loop_thread is not None:
                return

            self._close_event.clear()
            self._register_queue = queue.Queue()
            self._services = {}

            thread_target = partial(
                _start_zeroconf,
                self._close_event,
                self._services,
                self._services_lock,
                self._register_queue,
                self._address)

            self._zeroconf_loop_thread = threading.Thread(
                target=thread_target,
                daemon=True)

            self._zeroconf_loop_thread.start()

    @tornado.gen.coroutine
    def stop(self):
        """Signals the DNS-SD thread to stop and waits for the executor future to yield."""

        with (yield self._lock.acquire()):
            if self._zeroconf_loop_thread is None:
                return

            self._close_event.set()

            while self._zeroconf_loop_thread.is_alive():
                yield tornado.gen.sleep(ITER_WAIT)

            self._zeroconf_loop_thread = None
            self._close_event.clear()
            self._register_queue = None
            self._services = None

    @tornado.gen.coroutine
    def _run_task(self, task):
        """"""

        with (yield self._lock.acquire()):
            if self._zeroconf_loop_thread is None:
                raise ValueError("Stopped DNS-SD thread")

            done = threading.Event()
            task.update({'done': done})

            self._register_queue.put(task)

            while not done.is_set():
                yield tornado.gen.sleep(ITER_WAIT)

    @tornado.gen.coroutine
    def register(self, servient, instance_name=None):
        """Takes a Servient and registers the TD catalogue
        service for discovery by other hosts in the same link."""

        if instance_name and instance_name.endswith('.'):
            raise ValueError('Instance name ends with "."')

        yield self._run_task({
            'type': TASK_REGISTER,
            'servient': servient,
            'instance_name': instance_name
        })

    @tornado.gen.coroutine
    def unregister(self, servient, instance_name=None):
        """Takes a Servient and unregisters the TD catalogue service."""

        if instance_name and instance_name.endswith('.'):
            raise ValueError('Instance name ends with "."')

        yield self._run_task({
            'type': TASK_UNREGISTER,
            'servient': servient,
            'instance_name': instance_name
        })

    @tornado.gen.coroutine
    def find(self, min_results=None, timeout=5):
        """Browses the link to discover WoT Servient services using mDNS.
        Returns a list of (ip_address, port).
        If min_results is defined it will stop as soon as that number of results are found."""

        with (yield self._lock.acquire()):
            if self._zeroconf_loop_thread is None:
                raise ValueError("Stopped DNS-SD thread")

            ini = time.time()
            finished = False

            while (time.time() - ini) < timeout and not finished:
                if min_results is not None:
                    with self._services_lock:
                        if len(self._services) >= min_results:
                            finished = True

                yield tornado.gen.sleep(ITER_WAIT)

            with self._services_lock:
                found = list(six.itervalues(self._services))

            raise tornado.gen.Return(found)



import logging


class ConnRefCounter(object):
    """A simple connection reference counter to keep
    track of active connections and enable reuse."""

    def __init__(self):
        self._counter = {}
        self._logr = logging.getLogger(__name__)

    def increase(self, conn_id, ref_id):
        """Increases the reference counter for the connection."""

        if conn_id not in self._counter:
            self._counter[conn_id] = set()

        self._counter[conn_id].add(ref_id)

        self._logr.debug("Added ref {} to conn <{}> (current: {})".format(
            ref_id, conn_id, len(self._counter[conn_id])))

    def decrease(self, conn_id, ref_id):
        """Decreases the reference counter for the connection."""

        if conn_id not in self._counter:
            self._logr.warning("Attempted to decrease ref of unknown conn: {}".format(conn_id))
            return

        try:
            self._counter[conn_id].remove(ref_id)

            self._logr.debug("Removed ref {} from conn <{}> (current: {})".format(
                ref_id, conn_id, len(self._counter[conn_id])))
        except KeyError:
            self._logr.warning("Attempted to remove unknown reference: {}".format(ref_id))

    def has_any(self, conn_id):
        """Returns True if the connection has any references pointing to it."""

        return conn_id in self._counter and len(self._counter[conn_id])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that represents a group or set of ExposedThing instances that exist in the same context.
"""


class ExposedThingSet(object):
    """Represents a group of ExposedThing objects.
    A group cannot contain two ExposedThing with the same Thing ID."""

    def __init__(self):
        self._exposed_things = {}

    @property
    def exposed_things(self):
        """A generator that yields all the ExposedThing contained in this group."""

        for exposed_thing in self._exposed_things.values():
            yield exposed_thing

    def contains(self, exposed_thing):
        """Returns True if this group contains the given ExposedThing."""

        return exposed_thing in self._exposed_things.values()

    def add(self, exposed_thing):
        """Add a new ExposedThing to this set."""

        if exposed_thing.thing.id in self._exposed_things:
            raise ValueError("Duplicate Exposed Thing: {}".format(exposed_thing.name))

        self._exposed_things[exposed_thing.thing.id] = exposed_thing

    def remove(self, thing_id):
        """Removes an existing ExposedThing by ID.
        The thing_id argument may be the original Thing ID or the URL-safe name."""

        exposed_thing = self.find_by_thing_id(thing_id)

        if exposed_thing is None:
            raise ValueError("Unknown Exposed Thing: {}".format(thing_id))

        assert exposed_thing.thing.id in self._exposed_things
        self._exposed_things.pop(exposed_thing.thing.id)

    def find_by_thing_id(self, thing_id):
        """Finds an existing ExposedThing by Thing ID.
        The ID argument may be the original Thing ID or the URL-safe name
        (which is also unique and based on the ID)."""

        def is_match(exp_thing):
            return exp_thing.thing.id == thing_id or exp_thing.thing.url_name == thing_id

        return next((item for item in self._exposed_things.values() if is_match(item)), None)

    def find_by_interaction(self, interaction):
        """Finds the ExposedThing whose Thing contains the given Interaction."""

        def is_match(exp_thing):
            return exp_thing.thing is interaction.thing

        return next((item for item in self._exposed_things.values() if is_match(item)), None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
import time
import uuid

# noinspection PyPackageRequirements
import pytest
import tornado.concurrent
import tornado.gen
import tornado.gen
import tornado.ioloop
import tornado.ioloop
import tornado.testing
import tornado.websocket
import tornado.websocket
# noinspection PyPackageRequirements
from faker import Faker
from six.moves.urllib.parse import urlparse, urlunparse

from tests.utils import find_free_port
from wotpy.protocols.ws.server import WebsocketServer
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, ActionFragmentDict, EventFragmentDict
from wotpy.wot.exposed.thing import ExposedThing
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription
from wotpy.wot.thing import Thing


def build_websocket_url(exposed_thing, ws_server, server_port):
    """Returns the WS connection URL for the given ExposedThing."""

    base_url = ws_server.build_base_url(hostname="localhost", thing=exposed_thing.thing)
    parsed_url = urlparse(base_url)
    test_netloc = re.sub(r':(\d+)$', ':{}'.format(server_port), parsed_url.netloc)

    test_url_parts = list(parsed_url)
    test_url_parts[1] = test_netloc

    return urlunparse(test_url_parts)


@pytest.fixture
def websocket_server():
    """Builds a WebsocketServer instance with some ExposedThings."""

    servient = Servient()

    thing_01_id = uuid.uuid4().urn
    thing_02_id = uuid.uuid4().urn

    exposed_thing_01 = ExposedThing(servient=servient, thing=Thing(id=thing_01_id))
    exposed_thing_02 = ExposedThing(servient=servient, thing=Thing(id=thing_02_id))

    prop_name_01 = uuid.uuid4().hex
    prop_name_02 = uuid.uuid4().hex
    prop_name_03 = uuid.uuid4().hex
    event_name_01 = uuid.uuid4().hex
    action_name_01 = uuid.uuid4().hex

    prop_value_01 = Faker().sentence()
    prop_value_02 = Faker().sentence()
    prop_value_03 = Faker().sentence()

    prop_init_01 = PropertyFragmentDict({
        "type": "string",
        "observable": True
    })

    prop_init_02 = PropertyFragmentDict({
        "type": "string",
        "observable": True
    })

    prop_init_03 = PropertyFragmentDict({
        "type": "string",
        "observable": True
    })

    event_init_01 = EventFragmentDict({
        "type": "object"
    })

    action_init_01 = ActionFragmentDict({
        "input": {"type": "string"},
        "output": {"type": "string"}
    })

    def async_lower(parameters):
        loop = tornado.ioloop.IOLoop.current()
        input_value = parameters.get("input")
        return loop.run_in_executor(None, lambda x: time.sleep(0.1) or x.lower(), input_value)

    exposed_thing_01.add_property(prop_name_01, prop_init_01, value=prop_value_01)
    exposed_thing_01.add_property(prop_name_02, prop_init_02, value=prop_value_02)
    exposed_thing_01.add_event(event_name_01, event_init_01)
    exposed_thing_01.add_action(action_name_01, action_init_01, async_lower)

    exposed_thing_02.add_property(prop_name_03, prop_init_03, value=prop_value_03)

    ws_port = find_free_port()

    ws_server = WebsocketServer(port=ws_port)
    ws_server.add_exposed_thing(exposed_thing_01)
    ws_server.add_exposed_thing(exposed_thing_02)

    @tornado.gen.coroutine
    def start():
        yield ws_server.start()

    tornado.ioloop.IOLoop.current().run_sync(start)

    url_thing_01 = build_websocket_url(exposed_thing_01, ws_server, ws_port)
    url_thing_02 = build_websocket_url(exposed_thing_02, ws_server, ws_port)

    yield {
        "exposed_thing_01": exposed_thing_01,
        "exposed_thing_02": exposed_thing_02,
        "prop_name_01": prop_name_01,
        "prop_init_01": prop_init_01,
        "prop_value_01": prop_value_01,
        "prop_name_02": prop_name_02,
        "prop_init_02": prop_init_02,
        "prop_value_02": prop_value_02,
        "prop_name_03": prop_name_03,
        "prop_init_03": prop_init_03,
        "prop_value_03": prop_value_03,
        "event_name_01": event_name_01,
        "event_init_01": event_init_01,
        "action_name_01": action_name_01,
        "action_init_01": action_init_01,
        "ws_server": ws_server,
        "url_thing_01": url_thing_01,
        "url_thing_02": url_thing_02,
        "ws_port": ws_port
    }

    @tornado.gen.coroutine
    def stop():
        yield ws_server.stop()

    tornado.ioloop.IOLoop.current().run_sync(stop)


@pytest.fixture
def websocket_servient():
    """Returns a Servient that exposes a Websockets server and one ExposedThing."""

    ws_port = find_free_port()
    ws_server = WebsocketServer(port=ws_port)

    servient = Servient(catalogue_port=None)
    servient.add_server(ws_server)

    @tornado.gen.coroutine
    def start():
        raise tornado.gen.Return((yield servient.start()))

    wot = tornado.ioloop.IOLoop.current().run_sync(start)

    property_name_01 = uuid.uuid4().hex
    property_name_02 = uuid.uuid4().hex
    action_name_01 = uuid.uuid4().hex
    event_name_01 = uuid.uuid4().hex

    td_dict = {
        "id": uuid.uuid4().urn,
        "name": uuid.uuid4().hex,
        "properties": {
            property_name_01: {
                "observable": True,
                "type": "string"
            },
            property_name_02: {
                "observable": True,
                "type": "string"
            }
        },
        "actions": {
            action_name_01: {
                "input": {
                    "type": "object"
                },
                "output": {
                    "type": "string"
                },
            }
        },
        "events": {
            event_name_01: {
                "type": "string"
            }
        },
    }

    td = ThingDescription(td_dict)

    exposed_thing = wot.produce(td.to_str())
    exposed_thing.expose()

    @tornado.gen.coroutine
    def action_handler(parameters):
        input_value = parameters.get("input")
        arg_b = input_value.get("arg_b") or uuid.uuid4().hex
        raise tornado.gen.Return(input_value.get("arg_a") + arg_b)

    exposed_thing.set_action_handler(action_name_01, action_handler)

    yield servient

    @tornado.gen.coroutine
    def shutdown():
        yield servient.shutdown()

    tornado.ioloop.IOLoop.current().run_sync(shutdown)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Entities that handle the MQTT operations needed to support each of the Interaction verbs.

.. autosummary::
    :toctree: _handlers

    wotpy.protocols.mqtt.handlers.action
    wotpy.protocols.mqtt.handlers.base
    wotpy.protocols.mqtt.handlers.event
    wotpy.protocols.mqtt.handlers.ping
    wotpy.protocols.mqtt.handlers.property
    wotpy.protocols.mqtt.handlers.subs
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that contain the client logic for the Websocket protocol.
"""

import datetime
import logging
import uuid

import tornado.gen
import tornado.ioloop
import tornado.locks
import tornado.websocket
from rx import Observable
from tornado.concurrent import Future

from wotpy.protocols.client import BaseProtocolClient
from wotpy.protocols.enums import Protocols
from wotpy.protocols.exceptions import FormNotFoundException, ClientRequestTimeout
from wotpy.protocols.refs import ConnRefCounter
from wotpy.protocols.utils import pick_form, is_scheme_form
from wotpy.protocols.ws.enums import WebsocketMethods, WebsocketSchemes
from wotpy.protocols.ws.messages import \
    WebsocketMessageRequest, \
    WebsocketMessageResponse, \
    WebsocketMessageEmittedItem, \
    WebsocketMessageError, \
    WebsocketMessageException
from wotpy.wot.events import \
    PropertyChangeEmittedEvent, \
    EmittedEvent, \
    PropertyChangeEventInit


class WebsocketClient(BaseProtocolClient):
    """Implementation of the protocol client interface for the Websocket protocol."""

    SLEEP_AFTER_ERR_SECS = 1.0
    RECEIVE_LOOP_TERMINATE_SLEEP_SECS = 0.1

    def __init__(self, receive_timeout_secs=1.0, ping_interval=2000):
        self._receive_timeout_secs = receive_timeout_secs
        self._ping_interval = ping_interval
        self._conns = {}
        self._ref_counter = ConnRefCounter()
        self._lock_conn = tornado.locks.Lock()
        self._msg_conditions = {}
        self._messages = {}
        self._receive_stop_events = {}
        self._logr = logging.getLogger(__name__)

    @tornado.gen.coroutine
    def _init_conn(self, ws_url, ref_id):
        """Initializes and connects the WebSockets connection."""

        with (yield self._lock_conn.acquire()):
            self._ref_counter.increase(ws_url, ref_id)

            if ws_url in self._conns:
                return

            self._logr.debug("Connecting to <{}>".format(ws_url))

            self._conns[ws_url] = yield tornado.websocket.websocket_connect(
                ws_url,
                ping_interval=self._ping_interval)

            @tornado.gen.coroutine
            def _start_receive_loop():
                yield self._receive_loop(ws_url)

            self._receive_stop_events[ws_url] = tornado.locks.Event()
            tornado.ioloop.IOLoop.current().add_callback(_start_receive_loop)

    @tornado.gen.coroutine
    def _stop_conn(self, ws_url, ref_id):
        """Disconnects the WebSockets connection."""

        with (yield self._lock_conn.acquire()):
            self._ref_counter.decrease(ws_url, ref_id)

            if self._ref_counter.has_any(ws_url):
                return

            try:
                if ws_url in self._conns:
                    self._logr.debug("Disconnecting WS client: {}".format(ws_url))
                    yield self._conns[ws_url].close()
            except Exception as ex:
                self._logr.warning("Error disconnecting: {}".format(ex), exc_info=True)

            if ws_url in self._receive_stop_events:
                self._logr.debug("Stopping message read loop: {}".format(ws_url))

                self._receive_stop_events[ws_url].set()

                while self._receive_stop_events[ws_url].is_set():
                    yield tornado.gen.sleep(self.RECEIVE_LOOP_TERMINATE_SLEEP_SECS)

                self._receive_stop_events.pop(ws_url)

            self._conns.pop(ws_url, None)
            self._messages.pop(ws_url, None)
            self._msg_conditions.pop(ws_url, None)

    @tornado.gen.coroutine
    def _send_message(self, ws_url, msg_req):
        """Sends a WebSockets message and returns the condition
        that will be notified when the response arrives."""

        if ws_url not in self._conns:
            self._logr.warning("<{}> is not an active connection".format(ws_url))
            return

        if ws_url not in self._msg_conditions:
            self._msg_conditions[ws_url] = {}

        if msg_req.id in self._msg_conditions[ws_url]:
            self._logr.warning("Message condition already exists")

        yield self._conns[ws_url].write_message(msg_req.to_json())

        msg_condition = tornado.locks.Condition()
        self._msg_conditions[ws_url][msg_req.id] = msg_condition

        raise tornado.gen.Return(msg_condition)

    @tornado.gen.coroutine
    def _receive_loop(self, ws_url):
        """Starts the WebSockets message receiving loop."""

        if ws_url not in self._conns:
            self._logr.warning("<{}> is not an active connection".format(ws_url))
            return

        if ws_url not in self._messages:
            self._messages[ws_url] = {}

        while not self._receive_stop_events[ws_url].is_set():
            try:
                raw_res = yield self._conns[ws_url].read_message()

                self._logr.debug("Read message: {}".format(raw_res))

                if raw_res is None:
                    self._logr.debug("Cannot read message: Closed WS connection")
                    yield tornado.gen.sleep(self.SLEEP_AFTER_ERR_SECS)
                    continue

                msg_res = self._parse_msg_response(raw_res)

                if msg_res:
                    self._messages[ws_url][msg_res.id] = msg_res
                    conditions = self._msg_conditions.get(ws_url, None)

                    if conditions and msg_res.id in conditions:
                        self._logr.debug("Notifying: {}".format(msg_res.id))
                        conditions[msg_res.id].notify_all()
            except Exception as ex:
                self._logr.warning("Error in read loop: {}".format(ex), exc_info=True)
                yield tornado.gen.sleep(self.SLEEP_AFTER_ERR_SECS)

        self._receive_stop_events[ws_url].clear()

    @classmethod
    def _parse_msg_response(cls, raw_msg):
        """Returns a parsed WS Response message instance if
        the raw message format is valid and has the given ID.
        Raises Exception if the WS message is an error."""

        try:
            return WebsocketMessageResponse.from_raw(raw_msg)
        except WebsocketMessageException:
            pass

        try:
            return WebsocketMessageError.from_raw(raw_msg)
        except WebsocketMessageException:
            pass

        return None

    @classmethod
    def _parse_emitted_item(cls, raw_msg, sub_id):
        """Returns a parsed WS Emitted Item message instance if
        the raw message format is valid and has the given ID.
        Raises Exception if the WS message is an error."""

        try:
            msg = WebsocketMessageEmittedItem.from_raw(raw_msg)

            if msg.subscription_id == sub_id:
                return msg
        except WebsocketMessageException:
            pass

        try:
            err = WebsocketMessageError.from_raw(raw_msg)
            err_sub_id = err.data and err.data.get("subscription")

            if err_sub_id == sub_id:
                raise Exception(err.message)
        except WebsocketMessageException:
            pass

        return None

    @property
    def protocol(self):
        """Protocol of this client instance.
        A member of the Protocols enum."""

        return Protocols.WEBSOCKETS

    def _build_subscribe(self, ws_url, msg_req, on_next):
        """Builds the subscribe function that is passed
        as an argument on the creation of an Observable."""

        def subscribe(observer):
            """Connect to the WS server and start passing the received events to the Observer."""

            future_sub_id = Future()
            future_ws_conn = Future()

            def on_error(ex):
                observer.on_error(ex)

            def on_next_event(raw_msg):
                sub_id = future_sub_id.result()

                try:
                    msg_item = self._parse_emitted_item(raw_msg, sub_id)
                except Exception as ex:
                    return on_error(ex)

                if msg_item is None:
                    return

                try:
                    on_next(observer, msg_item)
                except Exception as ex:
                    return on_error(ex)

            def parse_subscription_id(raw_msg):
                msg_res = self._parse_msg_response(raw_msg)

                if isinstance(msg_res, WebsocketMessageError):
                    return on_error(Exception(msg_res.message))

                if msg_res is None:
                    return

                sub_id = msg_res.result
                future_sub_id.set_result(sub_id)

            def on_msg(raw_msg):
                if raw_msg is None:
                    return on_error(Exception("WS connection closed"))

                if future_sub_id.done():
                    on_next_event(raw_msg)
                else:
                    parse_subscription_id(raw_msg)

            def on_conn(ft):
                try:
                    ws_conn = ft.result()
                except Exception as ex:
                    return on_error(ex)

                future_ws_conn.set_result(ws_conn)
                ws_conn.write_message(msg_req.to_json())

            tornado.websocket.websocket_connect(ws_url, callback=on_conn, on_message_callback=on_msg)

            def unsubscribe():
                if future_ws_conn.done():
                    ws_conn = future_ws_conn.result()
                    ws_conn.close()

            return unsubscribe

        return subscribe

    def is_supported_interaction(self, td, name):
        """Returns True if the any of the Forms for the Interaction
        with the given name is supported in this Protocol Binding client."""

        forms = td.get_forms(name)

        forms_wss = [
            form for form in forms
            if is_scheme_form(form, td.base, WebsocketSchemes.WSS)
        ]

        forms_ws = [
            form for form in forms
            if is_scheme_form(form, td.base, WebsocketSchemes.WS)
        ]

        return len(forms_wss) or len(forms_ws)

    def _raise_message(self, ws_url, msg_id):
        """Raises the error or return Exception from the message
        in the internal collection matching the given ID."""

        assert ws_url in self._messages, "Unknown WS connection"
        assert msg_id in self._messages[ws_url], "Unknown message ID"

        msg = self._messages[ws_url][msg_id]

        if isinstance(msg, WebsocketMessageError):
            raise Exception(msg.message)
        else:
            raise tornado.gen.Return(msg.result)

    @tornado.gen.coroutine
    def invoke_action(self, td, name, input_value, timeout=None):
        """Invokes an Action on a remote Thing.
        Returns a Future."""

        if name not in td.actions:
            raise FormNotFoundException()

        form = pick_form(
            td, td.get_action_forms(name),
            WebsocketSchemes.list())

        if not form:
            raise FormNotFoundException()

        ws_url = form.resolve_uri(td.base)
        ref_id = uuid.uuid4().hex

        try:
            yield self._init_conn(ws_url, ref_id)

            msg_req = WebsocketMessageRequest(
                method=WebsocketMethods.INVOKE_ACTION,
                params={"name": name, "parameters": input_value},
                msg_id=uuid.uuid4().hex)

            condition = yield self._send_message(ws_url, msg_req)

            timeout = datetime.timedelta(seconds=timeout) if timeout else None
            cond_res = yield condition.wait(timeout=timeout)

            if not cond_res:
                raise ClientRequestTimeout

            self._raise_message(ws_url, msg_req.id)
        finally:
            yield self._stop_conn(ws_url, ref_id)

    @tornado.gen.coroutine
    def write_property(self, td, name, value, timeout=None):
        """Updates the value of a Property on a remote Thing.
        Returns a Future."""

        if name not in td.properties:
            raise FormNotFoundException()

        form = pick_form(
            td, td.get_property_forms(name),
            WebsocketSchemes.list())

        if not form:
            raise FormNotFoundException()

        ws_url = form.resolve_uri(td.base)
        ref_id = uuid.uuid4().hex

        try:
            yield self._init_conn(ws_url, ref_id)

            msg_req = WebsocketMessageRequest(
                method=WebsocketMethods.WRITE_PROPERTY,
                params={"name": name, "value": value},
                msg_id=uuid.uuid4().hex)

            condition = yield self._send_message(ws_url, msg_req)

            timeout = datetime.timedelta(seconds=timeout) if timeout else None
            cond_res = yield condition.wait(timeout=timeout)

            if not cond_res:
                raise ClientRequestTimeout

            self._raise_message(ws_url, msg_req.id)
        finally:
            yield self._stop_conn(ws_url, ref_id)

    @tornado.gen.coroutine
    def read_property(self, td, name, timeout=None):
        """Reads the value of a Property on a remote Thing.
        Returns a Future."""

        if name not in td.properties:
            raise FormNotFoundException()

        form = pick_form(
            td, td.get_property_forms(name),
            WebsocketSchemes.list())

        if not form:
            raise FormNotFoundException()

        ws_url = form.resolve_uri(td.base)
        ref_id = uuid.uuid4().hex

        try:
            yield self._init_conn(ws_url, ref_id)

            msg_req = WebsocketMessageRequest(
                method=WebsocketMethods.READ_PROPERTY,
                params={"name": name},
                msg_id=uuid.uuid4().hex)

            condition = yield self._send_message(ws_url, msg_req)

            timeout = datetime.timedelta(seconds=timeout) if timeout else None
            cond_res = yield condition.wait(timeout=timeout)

            if not cond_res:
                raise ClientRequestTimeout

            self._raise_message(ws_url, msg_req.id)
        finally:
            yield self._stop_conn(ws_url, ref_id)

    def on_event(self, td, name):
        """Subscribes to an event on a remote Thing.
        Returns an Observable."""

        if name not in td.events:
            # noinspection PyUnresolvedReferences
            return Observable.throw(FormNotFoundException())

        form = pick_form(
            td, td.get_event_forms(name),
            WebsocketSchemes.list())

        if not form:
            # noinspection PyUnresolvedReferences
            return Observable.throw(FormNotFoundException())

        ws_url = form.resolve_uri(td.base)

        msg_req = WebsocketMessageRequest(
            method=WebsocketMethods.ON_EVENT,
            params={"name": name},
            msg_id=uuid.uuid4().hex)

        def on_next(observer, msg_item):
            observer.on_next(EmittedEvent(init=msg_item.data, name=name))

        subscribe = self._build_subscribe(ws_url, msg_req, on_next)

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_property_change(self, td, name):
        """Subscribes to property changes on a remote Thing.
        Returns an Observable."""

        if name not in td.properties:
            # noinspection PyUnresolvedReferences
            return Observable.throw(FormNotFoundException())

        form = pick_form(
            td, td.get_property_forms(name),
            WebsocketSchemes.list())

        if not form:
            # noinspection PyUnresolvedReferences
            return Observable.throw(FormNotFoundException())

        ws_url = form.resolve_uri(td.base)

        msg_req = WebsocketMessageRequest(
            method=WebsocketMethods.ON_PROPERTY_CHANGE,
            params={"name": name},
            msg_id=uuid.uuid4().hex)

        def on_next(observer, msg_item):
            init_name = msg_item.data["name"]
            init_value = msg_item.data["value"]
            init = PropertyChangeEventInit(name=init_name, value=init_value)
            observer.on_next(PropertyChangeEmittedEvent(init=init))

        subscribe = self._build_subscribe(ws_url, msg_req, on_next)

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_td_change(self, url):
        """Subscribes to Thing Description changes on a remote Thing.
        Returns an Observable."""

        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that contain the client logic for the CoAP protocol.
"""

import asyncio
import json
import logging
import time

import aiocoap
import tornado.concurrent
import tornado.gen
import tornado.ioloop
import tornado.locks
from rx import Observable
from six.moves.urllib_parse import urlparse

from wotpy.protocols.client import BaseProtocolClient
from wotpy.protocols.coap.enums import CoAPSchemes
from wotpy.protocols.enums import Protocols, InteractionVerbs
from wotpy.protocols.exceptions import FormNotFoundException, ProtocolClientException, ClientRequestTimeout
from wotpy.protocols.utils import is_scheme_form
from wotpy.utils.utils import handle_observer_finalization
from wotpy.wot.events import PropertyChangeEventInit, PropertyChangeEmittedEvent, EmittedEvent


# noinspection PyCompatibility
class CoAPClient(BaseProtocolClient):
    """Implementation of the protocol client interface for the CoAP protocol."""

    def __init__(self):
        self._logr = logging.getLogger(__name__)
        self._coap_client = None
        self._client_lock = tornado.locks.Lock()
        super(CoAPClient, self).__init__()

    @classmethod
    def _pick_coap_href(cls, td, forms, op=None):
        """Picks the most appropriate CoAP form href from the given list of forms."""

        def is_op_form(form):
            try:
                return op is None or op == form.op or op in form.op
            except TypeError:
                return False

        def find_href(scheme):
            try:
                return next(
                    form.href for form in forms
                    if is_scheme_form(form, td.base, scheme) and is_op_form(form))
            except StopIteration:
                return None

        form_coaps = find_href(CoAPSchemes.COAPS)

        return form_coaps if form_coaps is not None else find_href(CoAPSchemes.COAP)

    @classmethod
    def _assert_success(cls, res):
        """Asserts that the given CoAP response was successful and raises an Exception if not."""

        if not res.code.is_successful():
            raise ProtocolClientException("Unsuccessful CoAP response: {}".format(res))

    def _build_subscribe(self, href, next_item_builder):
        """Builds the subscribe function that should be passed when
        constructing an Observable linked to an observable CoAP resurce."""

        def subscribe(observer):
            """Subscription function to observe resources using the CoAP protocol."""

            query = urlparse(href).query

            state = {
                "active": True,
                "request": None,
                "pending": None
            }

            @handle_observer_finalization(observer)
            @tornado.gen.coroutine
            def callback():
                self._logr.debug("Creating CoAP client for observation: {}".format(query))

                coap_client = yield aiocoap.Context.create_client_context()

                try:
                    msg = aiocoap.Message(code=aiocoap.Code.GET, uri=href, observe=0)
                    state["request"] = coap_client.request(msg)

                    self._logr.debug("Sending observation request: {}".format(msg))

                    future_first_resp = state["request"].response
                    state["pending"] = future_first_resp
                    first_resp = yield future_first_resp
                    state["pending"] = None
                    self._assert_success(first_resp)
                    next_item = next_item_builder(first_resp.payload)
                    next_item is not None and observer.on_next(next_item)

                    while state["active"]:
                        next_obsv_gen = state["request"].observation.__aiter__().__anext__()
                        future_resp = tornado.gen.convert_yielded(next_obsv_gen)
                        state["pending"] = future_resp
                        resp = yield future_resp
                        state["pending"] = None
                        self._assert_success(resp)
                        next_item = next_item_builder(resp.payload)
                        next_item is not None and observer.on_next(next_item)

                    self._logr.debug("Terminated subscription callback for: {}".format(query))
                finally:
                    yield coap_client.shutdown()

            def unsubscribe():
                self._logr.debug("Unsubscribing from: {}".format(query))

                state["active"] = False

                if state["request"] and not state["request"].observation.cancelled:
                    self._logr.debug("Cancelling observation on: {}".format(query))
                    state["request"].observation.cancel()

                if state["pending"]:
                    self._logr.debug("Cancelling pending request: {}".format(state["pending"]))
                    state["pending"].cancel()

            tornado.ioloop.IOLoop.current().add_callback(callback)

            return unsubscribe

        return subscribe

    @property
    def protocol(self):
        """Protocol of this client instance.
        A member of the Protocols enum."""

        return Protocols.COAP

    def is_supported_interaction(self, td, name):
        """Returns True if the any of the Forms for the Interaction
        with the given name is supported in this Protocol Binding client."""

        forms = td.get_forms(name)

        forms_coap = [
            form for form in forms
            if is_scheme_form(form, td.base, CoAPSchemes.list())
        ]

        return len(forms_coap) > 0

    async def _invocation_create(self, coap_client, href, input_value, timeout=None):
        """Creates a new action invocation by sending a POST request."""

        payload = json.dumps({"input": input_value}).encode("utf-8")
        msg = aiocoap.Message(code=aiocoap.Code.POST, payload=payload, uri=href)
        request = coap_client.request(msg)

        try:
            response = await asyncio.wait_for(request.response, timeout=timeout)
        except asyncio.TimeoutError:
            raise ClientRequestTimeout

        self._assert_success(response)

        invocation_id = json.loads(response.payload).get("id")

        return invocation_id

    async def _invocation_observe(self, coap_client, href, invocation_id, timeout=None):
        """Starts observing an existing action invocation by sending a GET request."""

        payload = json.dumps({"id": invocation_id}).encode("utf-8")
        msg = aiocoap.Message(code=aiocoap.Code.GET, payload=payload, uri=href, observe=0)
        request = coap_client.request(msg)

        try:
            response = await asyncio.wait_for(request.response, timeout=timeout)
        except asyncio.TimeoutError:
            raise ClientRequestTimeout

        self._assert_success(response)

        return request, response

    async def _invocation_next(self, request, timeout=None):
        """Waits for the next item in an active action invocation observation."""

        try:
            response = await asyncio.wait_for(
                request.observation.__aiter__().__anext__(),
                timeout=timeout)
        except asyncio.TimeoutError:
            raise ClientRequestTimeout

        self._assert_success(response)

        return response

    async def invoke_action(self, td, name, input_value, timeout=None):
        """Invokes an Action on a remote Thing."""

        href = self._pick_coap_href(
            td, td.get_action_forms(name),
            op=InteractionVerbs.INVOKE_ACTION)

        if href is None:
            raise FormNotFoundException()

        coap_client = await aiocoap.Context.create_client_context()

        try:
            invocation_id = await self._invocation_create(
                coap_client, href, input_value, timeout=timeout)

            request_obsv, response_obsv = await self._invocation_observe(
                coap_client, href, invocation_id, timeout=timeout)

            invocation_status = json.loads(response_obsv.payload)

            now = time.time()

            while not invocation_status.get("done"):
                if timeout and (time.time() - now) > timeout:
                    raise ClientRequestTimeout

                response_obsv = await self._invocation_next(request_obsv, timeout=timeout)
                invocation_status = json.loads(response_obsv.payload)

            if not request_obsv.observation.cancelled:
                request_obsv.observation.cancel()

            if invocation_status.get("error"):
                raise Exception(invocation_status.get("error"))
            else:
                return invocation_status.get("result")
        finally:
            await coap_client.shutdown()

    async def write_property(self, td, name, value, timeout=None):
        """Updates the value of a Property on a remote Thing."""

        href = self._pick_coap_href(
            td, td.get_property_forms(name),
            op=InteractionVerbs.WRITE_PROPERTY)

        if href is None:
            raise FormNotFoundException()

        coap_client = await aiocoap.Context.create_client_context()

        try:
            payload = json.dumps({"value": value}).encode("utf-8")
            msg = aiocoap.Message(code=aiocoap.Code.POST, payload=payload, uri=href)
            request = coap_client.request(msg)

            try:
                response = await asyncio.wait_for(request.response, timeout=timeout)
            except asyncio.TimeoutError:
                raise ClientRequestTimeout

            self._assert_success(response)
        finally:
            await coap_client.shutdown()

    async def read_property(self, td, name, timeout=None):
        """Reads the value of a Property on a remote Thing."""

        href = self._pick_coap_href(
            td, td.get_property_forms(name),
            op=InteractionVerbs.READ_PROPERTY)

        if href is None:
            raise FormNotFoundException()

        coap_client = await aiocoap.Context.create_client_context()

        try:
            msg = aiocoap.Message(code=aiocoap.Code.GET, uri=href)
            request = coap_client.request(msg)

            try:
                response = await asyncio.wait_for(request.response, timeout=timeout)
            except asyncio.TimeoutError:
                raise ClientRequestTimeout

            self._assert_success(response)

            prop_value = json.loads(response.payload).get("value")

            return prop_value
        finally:
            await coap_client.shutdown()

    def on_property_change(self, td, name):
        """Subscribes to property changes on a remote Thing.
        Returns an Observable"""

        href = self._pick_coap_href(
            td, td.get_property_forms(name),
            op=InteractionVerbs.OBSERVE_PROPERTY)

        if href is None:
            raise FormNotFoundException()

        def next_item_builder(payload):
            value = json.loads(payload).get("value")
            init = PropertyChangeEventInit(name=name, value=value)
            return PropertyChangeEmittedEvent(init=init)

        subscribe = self._build_subscribe(href, next_item_builder)

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_event(self, td, name):
        """Subscribes to an event on a remote Thing.
        Returns an Observable."""

        href = self._pick_coap_href(
            td, td.get_event_forms(name),
            op=InteractionVerbs.SUBSCRIBE_EVENT)

        if href is None:
            raise FormNotFoundException()

        def next_item_builder(payload):
            if payload:
                data = json.loads(payload).get("data")
                return EmittedEvent(init=data, name=name)
            else:
                return None

        subscribe = self._build_subscribe(href, next_item_builder)

        # noinspection PyUnresolvedReferences
        return Observable.create(subscribe)

    def on_td_change(self, url):
        """Subscribes to Thing Description changes on a remote Thing.
        Returns an Observable."""

        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Wrapper classes for security dictionaries defined in the Scripting API.
"""

from wotpy.wot.dictionaries.base import WotBaseDict
from wotpy.utils.utils import merge_args_kwargs_dict
from wotpy.wot.enums import SecuritySchemeType


class SecuritySchemeDict(WotBaseDict):
    """Contains security related configuration."""

    class Meta:
        fields = {
            "scheme",
            "description",
            "proxy"
        }

        required = {
            "scheme"
        }

    @classmethod
    def build(cls, *args, **kwargs):
        """Builds an instance of the appropriate subclass for the given SecurityScheme."""

        init_dict = merge_args_kwargs_dict(args, kwargs)

        klass_map = {
            SecuritySchemeType.NOSEC: NoSecuritySchemeDict,
            SecuritySchemeType.BASIC: BasicSecuritySchemeDict,
            SecuritySchemeType.CERT: CertSecuritySchemeDict,
            SecuritySchemeType.DIGEST: DigestSecuritySchemeDict,
            SecuritySchemeType.BEARER: BearerSecuritySchemeDict,
            SecuritySchemeType.PSK: PSKSecuritySchemeDict,
            SecuritySchemeType.PUBLIC: PublicSecuritySchemeDict,
            SecuritySchemeType.OAUTH2: OAuth2SecuritySchemeDict,
            SecuritySchemeType.APIKEY: APIKeySecuritySchemeDict,
            SecuritySchemeType.POP: PoPSecuritySchemeDict
        }

        scheme_type = init_dict.get("scheme")
        klass = klass_map.get(scheme_type)

        if not klass:
            raise ValueError("Unknown scheme: {}".format(scheme_type))

        return klass(*args, **kwargs)


class NoSecuritySchemeDict(SecuritySchemeDict):
    """A security configuration indicating there is no authentication
    or other mechanism required to access the resource."""

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.NOSEC


class BasicSecuritySchemeDict(SecuritySchemeDict):
    """Basic authentication security configuration using an unencrypted username and password."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "in",
            "name"
        })

        required = SecuritySchemeDict.Meta.required

        defaults = {
            "in": "header"
        }

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.BASIC


class CertSecuritySchemeDict(SecuritySchemeDict):
    """Certificate-base asymmetric key security configuration."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "identity"
        })

        required = SecuritySchemeDict.Meta.required

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.CERT


class DigestSecuritySchemeDict(SecuritySchemeDict):
    """Digest authentication security configuration. This scheme is similar to
    basic authentication but with added features to avoid man-in-the-middle attacks."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "qop",
            "in",
            "name"
        })

        required = SecuritySchemeDict.Meta.required

        defaults = {
            "qop": "auth",
            "in": "header"
        }

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.DIGEST


class BearerSecuritySchemeDict(SecuritySchemeDict):
    """Bearer token authentication security configuration. This scheme is intended
    for situations where bearer tokens are used independently of OAuth2.
    If the oauth2 scheme is specified it is not generally necessary to
    specify this scheme as well as it is implied."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "authorization",
            "alg",
            "format",
            "in",
            "name"
        })

        required = SecuritySchemeDict.Meta.required

        defaults = {
            "alg": "ES256",
            "format": "jwt",
            "in": "header"
        }

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.BEARER


class PSKSecuritySchemeDict(SecuritySchemeDict):
    """Pre-shared key authentication security configuration."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "identity"
        })

        required = SecuritySchemeDict.Meta.required

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.PSK


class PublicSecuritySchemeDict(SecuritySchemeDict):
    """Raw public key asymmetric key security configuration."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "identity"
        })

        required = SecuritySchemeDict.Meta.required

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.PUBLIC


class OAuth2SecuritySchemeDict(SecuritySchemeDict):
    """OAuth2 authentication security configuration.
    For the implicit flow the authorization and scopes are required.
    For the password and client flows both token and scopes are required.
    For the code flow authorization, token, and scopes are required."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "authorization",
            "token",
            "refresh",
            "scopes",
            "flow"
        })

        required = SecuritySchemeDict.Meta.required

        defaults = {
            "flow": "implicit"
        }

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.OAUTH2


class APIKeySecuritySchemeDict(SecuritySchemeDict):
    """API key authentication security configuration.
    This is for the case where the access token is opaque and is not using a standard token format."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "in",
            "name"
        })

        required = SecuritySchemeDict.Meta.required

        defaults = {
            "in": "query"
        }

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.APIKEY


class PoPSecuritySchemeDict(SecuritySchemeDict):
    """Proof-of-possession token authentication security configuration."""

    class Meta:
        fields = SecuritySchemeDict.Meta.fields.union({
            "alg",
            "authorization",
            "format",
            "in",
            "name"
        })

        required = SecuritySchemeDict.Meta.required

        defaults = {
            "alg": "ES256",
            "format": "jwt",
            "in": "header"
        }

    @property
    def scheme(self):
        """The scheme property represents the identification
        of the security scheme to be used for the Thing."""

        return SecuritySchemeType.POP
